import sys
import logging
import contextlib

logger = logging.getLogger(__name__)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setLevel(level=logging.DEBUG)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

@contextlib.contextmanager
def loglevel(level):
    levels = [handler.level for handler in logger.handlers]
    for handler in logger.handlers:
        handler.setLevel(level)
    yield
    for handler, olevel in zip(logger.handlers, levels):
        handler.setLevel(olevel)


from typing import Optional, List, Sequence, Any, Union
from nltk import ngrams
from pathlib import Path
import errno
import re
import os


class SpellSuggestions(object):
    def __init__(self, is_correct: bool = False, suggestions: Optional[List[str]] = None):
        self.is_correct = is_correct
        self.suggestions = suggestions or []


def dameraulevenshtein(seq1: Sequence[Any], seq2: Sequence[Any]):
    """
    Calculate the Damerau-Levenshtein distance between sequences.

    This method has not been modified from the original.
    Source: http://mwh.geek.nz/2009/04/26/python-damerau-levenshtein-distance/

    This distance is the number of additions, deletions, substitutions,
    and transpositions needed to transform the first sequence into the
    second. Although generally used with strings, any sequences of
    comparable objects will work.

    Transpositions are exchanges of *consecutive* characters; all other
    operations are self-explanatory.

    This implementation is O(N*M) time and O(M) space, for N and M the
    lengths of the two sequences.

    >>> dameraulevenshtein('ba', 'abc')
    2
    >>> dameraulevenshtein('fee', 'deed')
    2

    It works with arbitrary sequences too:
    >>> dameraulevenshtein('abcd', ['b', 'a', 'c', 'd', 'e'])
    2

    Args:
       seq1 (str): word 1
       seq2 (str): word 2

    Returns:
        edit_distance (int): Damerau-Levenshtein distance
    """
    # Conceptually, this is based on a len(seq1) + 1 * len(seq2) + 1 matrix.
    # However, only the current and two previous rows are needed at once,
    # so we only store those.
    oneago = None
    thisrow = []
    for num in range(1, len(seq2) + 1):
        thisrow.append(num)
    thisrow += [0]
    for x in range(len(seq1)):
        # Python lists wrap around for negative indices, so put the
        # leftmost column at the *end* of the list. This matches with
        # the zero-indexed strings and saves extra calculation.
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in range(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)
            # This block deals with transpositions
            if x > 0 and y > 0 and seq1[x] == seq2[y - 1] and seq1[x - 1] == seq2[y] and seq1[x] != seq2[y]:
                thisrow[y] = min(thisrow[y], twoago[y - 2] + 1)
    return thisrow[len(seq2) - 1]


def batch(iterable, n=1):
    iter_len = len(iterable)
    for ndx in range(0, iter_len, n):
        yield iterable[ndx:min(ndx + n, iter_len)]


def get_ngrams(tokens, n):
    return list(ngrams(tokens, n))


def get_clean_text(sentence):
    clean_text = re.sub(r'[.,:;\"?\\]', ' ', str(sentence))
    clean_text = re.sub(r'{.*}', '', clean_text)
    return clean_text.strip()


def mkdirs(path: Union[str, Path], err_if_already_exists: bool = True) -> None:
    """
    Given a path to a directory, create the directory and all intermediate directories if they don't exist

    Args:
        path: path to some directory
        err_if_already_exists: whether to throw error if directory already exists

    Returns:
        None

    Raises:
        OSError: If the directory already exists

    """
    path = str(path)
    try:
        os.makedirs(path)
    except FileExistsError:
        if err_if_already_exists:
            raise
    except OSError as e:
        if e.errno == errno.EEXIST:
            if err_if_already_exists:
                raise
        else:
            raise



from collections import defaultdict
from itertools import repeat

from spello.utils import dameraulevenshtein, SpellSuggestions
from spello.phoneme.constant import (
    indic_char_map,
    indic_language_chars
)
import operator


class PhonemeModel(object):
    indic_languages = list(indic_language_chars.keys())
    supported_languages = indic_languages + ['en']

    def __init__(self, config, script, phoneme_dict=None):
        self.phoneme_dict = phoneme_dict
        self.script = script
        self.config = config

    @staticmethod
    def get_latin_soundex(token):
        """
        Get the phoneme code for the latin word
        Args:
            token: word
        Returns:
            (str): soundex code
        Examples:
            >>> token = 'availability'
            >>> PhonemeModel().get_latin_soundex(token)
            'A181870000'
        """
        token = token.upper()

        soundex = ""

        # first letter of input is always the first letter of phoneme
        soundex += token[0]

        # create a dictionary which maps letters to respective phoneme codes. Vowels and 'H', 'W' and 'Y' will be
        # represented by '.'
        dictionary = {"BPV": "1", "F": "2", "C": "3", "GJ": "4", "KQ": "5", "SXZ": "6", "DT": "7", "L": "8",
                      "MN": "9", "R": "10", "AEIOUHWY": "."}

        for char in token[1:]:
            for key in dictionary.keys():
                if char in key:
                    code = dictionary[key]
                    if code != soundex[-1]:
                        soundex += code

        # remove vowels and 'H', 'W' and 'Y' from phoneme
        soundex = soundex.replace(".", "")

        # trim or pad to make phoneme a 10-character code
        soundex = soundex[:10].ljust(10, "0")

        return soundex

    def get_indic_soundex(self, token, length=8):
        """
        Get Soundex code for indic languages
        Args:
            token (word): word
            length (int): length of soundex code
        Returns:
            (str): soundex code
        Examples:
            >>> token = 'उपलब्धता'
            >>> PhonemeModel().get_latin_soundex(token)
            'उMQMKA00'

        """
        sndx = []
        fc = token[0]

        # translate alpha chars in name to phoneme digits
        for char in token[1:].lower():
            if char not in indic_language_chars[self.script]:
                continue
            char_sound = str(indic_char_map[indic_language_chars[self.script].index(char)])

            # remove all 0s from the phoneme code
            if char_sound == '0':
                continue

            # duplicate consecutive phoneme digits are skipped
            if len(sndx) == 0:
                sndx.append(char_sound)
            elif char_sound != sndx[-1]:
                sndx.append(char_sound)

        # append first character to result
        sndx.insert(0, fc)

        if len(sndx) < length:
            sndx.extend(repeat('0', length))
            return ''.join(sndx[:length])

        return ''.join(sndx[:length])

    def get_soundex(self, word):
        """
        Return soundex code of word
        Args:
            word (str): word
        Returns:
            (str): soundex code
        """
        if self.script not in PhonemeModel.supported_languages:
            return None
        if self.script == "en":
            return self.get_latin_soundex(word)
        else:
            return self.get_indic_soundex(word)

    def create_phoneme_dictionary_from_sentences(self, sentences):
        """
        Create phoneme dictionary from sentences.
        - get all tokens from sentences
        - find soundex code for each token
        - create dict with
            - key: soundex code
            - value : list of words belong to key (soundex code)
        Args:
            sentences (list): list of sentences
        Returns:
            (dict): phoneme dict
        """
        phoneme_dict = defaultdict(set)
        words = " ".join(sentences).strip().split()

        words_soundex = {word: self.get_soundex(word) for word in words if len(word) > 2}
        for word, soundex in words_soundex.items():
            phoneme_dict[soundex].add(word)
        self.phoneme_dict = phoneme_dict
        return self.phoneme_dict

    def create_phoneme_dictionary_from_words(self, words_counter):
        """
        Create phoneme dict from words counter (having dict of word and their count)
        follow below steps to create dict:
        - get all tokens from words counter
        - find soundex code for each token
        - create dict with
            - key: soundex code
            - value : list of words belong to key (soundex code)
        Args:
            words_counter (dict): word and their count
        Returns:
            (dict): phoneme dict
        """
        phoneme_dict = defaultdict(set)
        words_soundex = {word: self.get_soundex(word) for word in words_counter if len(word) > 2}
        for word, soundex in words_soundex.items():
            phoneme_dict[soundex].add(word)
        self.phoneme_dict = phoneme_dict
        return self.phoneme_dict

    def spell_correct(self, word) -> SpellSuggestions:
        """
        Suggest words from same soundex code
        Steps are:
            - Find soundex code of word
            - Return None if
                - no soundex of word is found
                - or word soundex not exists in phoneme dict
                - or word exists in word soundex list
            - else:
                - get all words of same soundex code from phoneme dict

                - if len(word) == 3
                    - allow all 1 edit word
                    - allow 2 edit word if suggested word and mispelled word ends with same character
                      (by default all suggestions will start with same character)

                if 4 <= len(word) <= 5:
                    - allow all words upto 2 edit distance
                    - allow 3 edit word if suggested word
                      same character as of word

                if 6 <= len(word) <= 9:
                    - allow all words up to 3 edit distance

                if len(word) >= 10:
                    - allow all words up to 4 edit distance
        Args:
            word (str): word to be corrected
        Returns:
            (list): list of tuples of word with their edit distances

        Examples:
            >>> word = 'apple'
            >>> PhonemeModel().spell_correct(word)
            SpellSuggestions(is_correct=False, suggestions=[('aple', 1), ('aaple', 1)])
        """

        spell_suggestions = SpellSuggestions(is_correct=False, suggestions=[])
        word_soundex = self.get_soundex(word)
        if not word_soundex or (word_soundex not in self.phoneme_dict):
            return spell_suggestions

        if word in self.phoneme_dict[word_soundex]:
            spell_suggestions.is_correct = True
            return spell_suggestions

        possible_suggestions = self.phoneme_dict[word_soundex]
        distances = [dameraulevenshtein(word, s_word) for s_word in possible_suggestions]

        # if script belongs to indic language, don't filter suggestion based on edit distance
        if self.script in PhonemeModel.indic_languages:
            spell_suggestions.suggestions = [(word, distances[index])
                                             for index, word in enumerate(possible_suggestions)]
            return spell_suggestions

        suggestions = []
        word_len = len(word)
        edit_distance_allowed = self.config.phoneme_allowed_distance_map[word_len]
        for index, s_word in enumerate(possible_suggestions):
            if distances[index] <= edit_distance_allowed:
                suggestions.append((s_word, distances[index]))

            if self.config.allow_1_extra_edit_for_char_start:
                edit_distance = edit_distance_allowed + 1
                if s_word not in suggestions and s_word[0] == word[0] and distances[index] <= edit_distance:
                    suggestions.append((s_word, distances[index]))

            if self.config.allow_1_extra_edit_for_char_end:
                edit_distance = edit_distance_allowed + 1
                if s_word not in suggestions and s_word[-1] == word[-1] and distances[index] <= edit_distance:
                    suggestions.append((s_word, distances[index]))

        suggestions.sort(key=operator.itemgetter(1))
        spell_suggestions.suggestions = suggestions
        return spell_suggestions




SYMSPELL_ALLOWED_DISTANCES_MAP = {
    3: 1,
    4: 1,
    5: 1,
    6: 2,
    7: 2,
    8: 2,
    9: 3,
    10: 3,
    11: 3,
    12: 3,
    13: 3,
    14: 3,
    15: 3
}

PHONEME_ALLOWED_DISTANCES_MAP = {
    3: 1,
    4: 2,
    5: 2,
    6: 3,
    7: 3,
    8: 3,
    9: 3,
    10: 3,
    11: 4,
    12: 4,
    13: 4,
    14: 4,
    15: 4
}


class Config(object):
    max_length_for_spellcorrection: int = 15
    min_length_for_spellcorrection: int = 3
    allow_1_extra_edit_for_char_end: bool = True
    allow_1_extra_edit_for_char_start: bool = True
    symspell_allowed_distance_map = SYMSPELL_ALLOWED_DISTANCES_MAP
    phoneme_allowed_distance_map = PHONEME_ALLOWED_DISTANCES_MAP



import re
import time
import logging

from spello.utils import dameraulevenshtein, SpellSuggestions

spellcorrection_logger = logging.getLogger('spellcorrection')


class SymSpell:
    def __init__(self, config, max_edit_distance=3, verbose=1, script="en"):
        """
        SymSpell is Symmetric Delete spelling correction algorithm

        SymSpell Algorithm resources :-
        https://github.com/wolfgarbe/SymSpell
        https://medium.com/@wolfgarbe/1000x-faster-spelling-correction-algorithm-2012-8701fcd87a5f

        Generate terms with an edit distance (deletes only) from each dictionary term and add them together
        with the original term to the dictionary. This is done only during model creation.
        Generate terms with an edit distance (deletes only) from the input term and searches them in the dictionary.

        It is language independent since it relies only on deletes instead of transposes + replaces + inserts + deletes

        Args:
            max_edit_distance (int): max edit distance
            verbose (int): verbose level
            script (str): language script
        """
        self.max_edit_distance = max_edit_distance
        self.verbose = verbose
        # 0: top suggestion
        # 1: all suggestions of smallest edit distance
        # 2: all suggestions <= max_edit_distance (slower, no early termination)

        self.dictionary = {}
        self.longest_word_length = 0
        self.script = script
        self.config = config

    @staticmethod
    def get_deletes_list(w):
        """
        Given a word, derive strings with up to max_edit_distance characters deleted

        Args:
            w (str): word for which we need to derive strings with up to max_edit_distance characters deleted

        Returns:
            deletes (list): list of words
        """
        deletes = []
        queue = [w]

        if len(w) <= 2:
            return [w]
        elif len(w) == 3:
            max_edit_distance = 1
        elif 4 <= len(w) < 6:
            max_edit_distance = 2
        elif 6 <= len(w) < 15:
            max_edit_distance = 3
        else:
            return [w]
        for d in range(max_edit_distance):
            temp_queue = []
            for word in queue:
                if len(word) > 1:
                    for c in range(len(word)):  # character index
                        word_minus_c = word[:c] + word[c + 1:]
                        if word_minus_c not in deletes:
                            deletes.append(word_minus_c)
                        if word_minus_c not in temp_queue:
                            temp_queue.append(word_minus_c)
            queue = temp_queue
        return deletes

    def create_dictionary_entry(self, w, count=None):
        """
        Add word and its derived deletions to dictionary

        Args:
            w (str): word for which we need to derive strings with up to max_edit_distance characters deleted
            count (int): count of word in corpus
        Returns:
            new_real_word_added (bool): whether word was added to dictionary i.e. first appearance of word in corpus
        """
        # check if word is already in dictionary
        # dictionary entries are in the form: (list of suggested corrections,
        # frequency of word in corpus)
        new_real_word_added = False
        if w in self.dictionary:
            # increment count of word in corpus
            if count:
                self.dictionary[w] = (self.dictionary[w][0], count)
            else:
                self.dictionary[w] = (self.dictionary[w][0], self.dictionary[w][1] + 1)
        else:
            self.dictionary[w] = ([], count)
            self.longest_word_length = max(self.longest_word_length, len(w))

        if self.dictionary[w][1] > 0:
            # first appearance of word in corpus
            # n.b. word may already be in dictionary as a derived word
            # (deleting character from a real word)
            # but counter of frequency of word in corpus is not incremented
            # in those cases)
            new_real_word_added = True
            deletes = self.get_deletes_list(w)
            for item in deletes:
                if item in self.dictionary:
                    # add (correct) word to delete's suggested correction list
                    self.dictionary[item][0].append(w)
                else:
                    # note frequency of word in corpus is not incremented
                    self.dictionary[item] = ([w], 0)

        return new_real_word_added

    def create_dictionary_from_sentences(self, lines):
        """
        Create dictionary from list of sentences

        Args:
            lines (list): list of sentences

        Returns:
            dictionary (dict): spell check dictionary
        """
        total_word_count = 0
        unique_word_count = 0
        start_time = time.time()
        spellcorrection_logger.info("Creating spell check dictionary...")
        punctuation = r""",+:?!"()!'.%[]"""
        pattern = re.compile(re.escape(punctuation))
        for line in lines:
            line = pattern.sub(' ', line.lower().strip())
            words = line.lower().strip().split()
            for word in words:
                total_word_count += 1
                if self.create_dictionary_entry(word):
                    unique_word_count += 1
        run_time = time.time() - start_time
        spellcorrection_logger.info("%.2f seconds to run" % run_time)
        spellcorrection_logger.info("total words processed: %i" % total_word_count)
        spellcorrection_logger.info("total unique words in corpus: %i" % unique_word_count)
        spellcorrection_logger.info("total items in dictionary (corpus words & deletions): %i" % len(self.dictionary))
        spellcorrection_logger.info("edit distance for deletions: %i" % self.max_edit_distance)
        spellcorrection_logger.info("length of longest word in corpus: %i" % self.longest_word_length)
        return self.dictionary

    def create_dictionary_from_words(self, words_counter):
        """
        Create dictionary from list of sentences

        Args:
            words_counter (dict): dict with word and their count

        Returns:
            dictionary (dict): spell check dictionary
        """
        total_word_count = 0
        unique_word_count = 0
        start_time = time.time()
        spellcorrection_logger.info("Creating spell check dictionary...")

        for word, count in words_counter.items():
            total_word_count += 1
            if self.create_dictionary_entry(word, count):
                unique_word_count += 1
        run_time = time.time() - start_time
        spellcorrection_logger.info("%.2f seconds to run" % run_time)
        spellcorrection_logger.info("total words processed: %i" % total_word_count)
        spellcorrection_logger.info("total unique words in corpus: %i" % unique_word_count)
        spellcorrection_logger.info("total items in dictionary (corpus words & deletions): %i" % len(self.dictionary))
        spellcorrection_logger.info("edit distance for deletions: %i" % self.max_edit_distance)
        spellcorrection_logger.info("length of longest word in corpus: %i" % self.longest_word_length)
        return self.dictionary

    def get_suggestions(self, string, silent=False):
        """
        Return list of suggested corrections for potentially incorrectly spelled word

        Option 1:
        ['file', 'five', 'fire', 'fine', ...]

        Option 2:
        [('file', (5, 0)),
         ('five', (67, 1)),
         ('fire', (54, 1)),
         ('fine', (17, 1))...]

        Args:
            string (str): word which needs to be corrected
            silent (bool): if silent is True, then logs won't be printed

        Returns:
            outlist (list): output list
        """
        if (len(string) - self.longest_word_length) > self.max_edit_distance:
            if not silent:
                spellcorrection_logger.info("no items in dictionary within maximum edit distance")
            return []

        suggest_dict = {}
        min_suggest_len = float('inf')

        queue = [string]
        q_dictionary = {}  # items other than string that we've checked

        while len(queue) > 0:
            q_item = queue[0]  # pop
            queue = queue[1:]

            # early exit
            if ((self.verbose < 2) and (len(suggest_dict) > 0) and
                    ((len(string) - len(q_item)) > min_suggest_len)):
                break

            # process queue item
            if (q_item in self.dictionary) and (q_item not in suggest_dict):
                if self.dictionary[q_item][1] > 0:
                    # word is in dictionary, and is a word from the corpus, and
                    # not already in suggestion list so add to suggestion
                    # dictionary, indexed by the word with value (frequency in
                    # corpus, edit distance)
                    # note q_items that are not the input string are shorter
                    # than input string since only deletes are added (unless
                    # manual dictionary corrections are added)
                    assert len(string) >= len(q_item)
                    suggest_dict[q_item] = (self.dictionary[q_item][1],
                                            len(string) - len(q_item))
                    # early exit
                    if (self.verbose < 2) and (len(string) == len(q_item)):
                        break
                    elif (len(string) - len(q_item)) < min_suggest_len:
                        min_suggest_len = len(string) - len(q_item)

                # the suggested corrections for q_item as stored in
                # dictionary (whether or not q_item itself is a valid word
                # or merely a delete) can be valid corrections
                for sc_item in self.dictionary[q_item][0]:
                    if sc_item not in suggest_dict:

                        # compute edit distance
                        # suggested items should always be longer
                        # (unless manual corrections are added)
                        assert len(sc_item) > len(q_item)

                        # q_items that are not input should be shorter
                        # than original string
                        # (unless manual corrections added)
                        assert len(q_item) <= len(string)

                        if len(q_item) == len(string):
                            assert q_item == string
                            # item_dist = len(sc_item) - len(q_item)

                        # item in suggestions list should not be the same as
                        # the string itself
                        assert sc_item != string

                        # calculate edit distance using, for example,
                        # Damerau-Levenshtein distance
                        item_dist = dameraulevenshtein(sc_item, string)

                        # do not add words with greater edit distance if
                        # verbose setting not on
                        if (self.verbose < 2) and (item_dist > min_suggest_len):
                            pass
                        elif item_dist <= self.max_edit_distance:
                            assert sc_item in self.dictionary  # should already be in dictionary if in suggestion list
                            suggest_dict[sc_item] = (self.dictionary[sc_item][1], item_dist)
                            if item_dist < min_suggest_len:
                                min_suggest_len = item_dist

                        # depending on order words are processed, some words
                        # with different edit distances may be entered into
                        # suggestions; trim suggestion dictionary if verbose
                        # setting not on
                        if self.verbose < 2:
                            suggest_dict = {k: v for k, v in suggest_dict.items() if v[1] <= min_suggest_len}

            # now generate deletes (e.g. a substring of string or of a delete)
            # from the queue item
            # as additional items to check -- add to end of queue
            assert len(string) >= len(q_item)

            # do not add words with greater edit distance if verbose setting
            # is not on
            if (self.verbose < 2) and ((len(string) - len(q_item)) > min_suggest_len):
                pass
            elif (len(string) - len(q_item)) < self.max_edit_distance and len(q_item) > 1:
                for c in range(len(q_item)):  # character index
                    word_minus_c = q_item[:c] + q_item[c + 1:]
                    if word_minus_c not in q_dictionary:
                        queue.append(word_minus_c)
                        q_dictionary[word_minus_c] = None  # arbitrary value, just to identify we checked this

        # queue is now empty: convert suggestions in dictionary to
        # list for output
        if not silent and self.verbose != 0:
            spellcorrection_logger.info("number of possible corrections: %i" % len(suggest_dict))
            spellcorrection_logger.info("edit distance for deletions: %i" % self.max_edit_distance)

        # output option 1
        # sort results by ascending order of edit distance and descending
        # order of frequency
        #     and return list of suggested word corrections only:
        # return sorted(suggest_dict, key = lambda x:
        #               (suggest_dict[x][1], -suggest_dict[x][0]))

        # output option 2
        # return list of suggestions with (correction,
        #                                  (frequency in corpus, edit distance)):
        as_list = list(suggest_dict.items())
        # e.g. [('a', (3, 1)), ('b', (2, 1)), ('c', (5, 2))]
        # outlist = sorted(as_list, key=lambda (term, (freq, dist)): (dist, -freq))
        outlist = sorted(as_list, key=lambda suggestion_tuple: (suggestion_tuple[1][1], -suggestion_tuple[1][0]))

        if self.verbose == 0:
            return outlist[0]
        else:
            return outlist

    def load_dictionary_data(self, dictionary_data):
        """
        Load dictionary data from dict

        Args:
            dictionary_data (dict): dict containing max_edit_distance, longest_word_length, dictionary and script keys

        Returns:
            None
        """
        if not dictionary_data:
            raise ValueError("dictionary_data is empty")

        if type(dictionary_data) != dict:
            raise TypeError("dictionary_data should be a dict")

        self.max_edit_distance = dictionary_data["max_edit_distance"]
        self.longest_word_length = dictionary_data["longest_word_length"]
        self.dictionary = dictionary_data["dictionary"]
        self.script = dictionary_data["script"]

    def get_dictionary_data(self):
        """
        Get dictionary data

        Returns:
            dictionary_data (dict): dict containing max_edit_distance, longest_word_length, dictionary and script keys

        """
        dictionary_data = {
            "max_edit_distance": self.max_edit_distance,
            "longest_word_length": self.longest_word_length,
            "dictionary": self.dictionary,
            "script": self.script
        }
        return dictionary_data

    def spell_correct(self, word):
        """
        Suggesting word from dictionary using below logic
        - if word exists in dictionary having count > 0 , return None
        - else
            - get suggestions list

                - if len(word) == 3,
                   - allow only 1 edit distance words

                - if 4 <= len(word) <= 5,
                    - allow all 1 edit distance words
                    - allow 2 edit word
                        - if first and last character of word and suggested word are same
                        - first letter of both are same

                - if 6 <= len(word) <=8:
                    - allow all upto 2 edit distance
                    - allow 3 edit word
                      - if first and last character of word and suggested word are same
                      - first letter of both are same

                - if len(word) >= 9:
                    - allow all word upto 3 edit distance
        Args:
            word (str): word to be corrected
        Returns:
            (SpellSuggestions): Spell suggestions instance with keys
                                is_correct: whether given spelling is correct
                                suggestions: List of tuples with words and their edit distance

        Examples:
            >>> word = 'ablee'
            >>> SymSpell().spell_correct(word)
            SpellSuggestions(is_correct=False, suggestions=[('able', 1), ('bleed', 2)])
        """

        spell_suggestions = SpellSuggestions(is_correct=False, suggestions=[])

        value = self.dictionary.get(word)
        if value and value[1] > 0:
            spell_suggestions.is_correct = True
            return spell_suggestions

        _suggestions = self.get_suggestions(word)
        if _suggestions:
            word_len = len(word)
            edit_distance_allowed = self.config.symspell_allowed_distance_map[word_len]
            suggestions = [(w, meta[1]) for w, meta in _suggestions if meta[1] <= edit_distance_allowed]

            if self.config.allow_1_extra_edit_for_char_end:
                edit_distance = edit_distance_allowed + 1
                for w, meta in _suggestions:
                    if w not in suggestions and w[0] == word[0] and meta[1] <= edit_distance:
                        suggestions.append((w, meta[1]))

            if self.config.allow_1_extra_edit_for_char_start:
                edit_distance = edit_distance_allowed + 1
                for w, meta in _suggestions:
                    if w not in suggestions and w[-1] == word[-1] and meta[1] <= edit_distance:
                        suggestions.append((w, meta[1]))

            spell_suggestions.suggestions = suggestions

        return spell_suggestions



from collections import defaultdict
from itertools import combinations
from typing import Dict, List, Tuple, NamedTuple, Union

from spello.utils import get_ngrams
import re

MAX_COUNT_ALLOWED = 100


class MemoryItem(NamedTuple):
    score: float
    decoded: Tuple[int, ...]


def get_context_pairs(tokens):
    """
    For given list of tokens, return all pairs possible within window of 4 words, ignore single letter word
    Args:
        tokens (list): list of tokens
    Returns:
        (list): list of pairs with their distances in token list
    Examples:
        >>> tokens
        >>> get_context_pairs(["please", "book", "a", "flight", "for", "me", "to", " pune"])
        {((' pune', 'for'), 3),
         ((' pune', 'me'), 2),
         ((' pune', 'to'), 1),
         (('book', 'flight'), 2),
         (('book', 'for'), 3),
         (('book', 'please'), 1),
         (('flight', 'for'), 1),
         (('flight', 'me'), 2),
         (('flight', 'please'), 3),
         (('flight', 'to'), 3),
         (('for', 'me'), 1),
         (('for', 'to'), 2),
         (('me', 'to'), 1)}
    """
    data = set()
    ngrams = get_ngrams(tokens, 4)
    if not ngrams:
        ngrams = [tokens]
    for ngrams_batch in ngrams:
        for pair in combinations(ngrams_batch, 2):
            diff_index = abs(tokens.index(pair[0]) - abs(tokens.index(pair[1])))
            if len(pair[0]) < 2 or len(pair[1]) < 2:
                continue
            data.add((pair, diff_index))
    return data


class ContextModel(object):
    """
    Context Model to suggest most suitable word in a sentence from given list of suggested words
    """
    START_TOKENS = ['^^', '^']
    END_TOKENS = ['$']

    def __init__(self):
        self.default_prob = None
        self.model_dict_count = None
        self.model_dict = None

    @staticmethod
    def get_corrected_words_map(correct_sentence: str, original_sentence: str) -> Dict[str, str]:
        """
        Create a dict having mapping of original words and their changed form in corrected sentence
        Args:
            correct_sentence (str): sentence formed after context suggestion
            original_sentence (str): original sentence
        Returns:
            (dict):
        """
        words_map = {}
        correct_words = correct_sentence.split()
        original_words = original_sentence.split()

        for correct_word, original_word in zip(correct_words, original_words):
            if correct_word != original_word.lower():
                words_map[original_word] = correct_word
        return words_map

    def create_model_dict(
            self,
            sentences: List[str],
    ) -> Dict[Tuple[str, ...], float]:

        """
        Create context model dict from given list of sentences
        steps followed are:
         - for each sentence find all context pairs possible
         - create dict having context pairs and their count
         - create dict from above to convert them into probabilities
        Args:
            sentences (list): list of sentences
        Returns:
            (dict): model dict
        """
        sentences_tokens = [
            ContextModel.START_TOKENS + str(sentence).lower().strip().split() + ContextModel.END_TOKENS
            for sentence in sentences
        ]
        model_dict_count: Dict[Tuple[str, ...], Union[float, int]] = defaultdict(int)
        model_dict: Dict[Tuple[str, ...], float] = defaultdict(float)

        for sentence_tokens in sentences_tokens:
            tokens = sentence_tokens
            context_pairs = get_context_pairs(tokens)
            for (pair, diff) in context_pairs:
                model_dict_count[pair] = min(model_dict_count[pair] + 1, MAX_COUNT_ALLOWED)

        values_model_dict_count = model_dict_count.values()
        total_count = float(sum(values_model_dict_count))
        self.default_prob = (min(values_model_dict_count) / total_count) * 0.5
        # to ensure self.default_prob is smaller than the smallest probability.

        for tup in model_dict_count:
            model_dict[tup] = model_dict_count[tup] / total_count

        self.model_dict = dict(model_dict)
        self.model_dict_count = dict(model_dict_count)

        return self.model_dict

    def get_most_probable_sentence(
            self,
            suggestions: List[List[str]]
    ) -> str:
        """
        Find most probable sentence from given list of sentences that can be formed trying all combination of
        words from suggestion
        Uses context model dict to find overall probability of formed sentence and return sentence with max
        probability

        Args:
            suggestions (List[List[str]]): List of lists representation of all probable sentences.

        Returns:
            (str): Spell corrected sentence.
        """
        sent_word_count = len(suggestions)
        suggestions = [[tok] for tok in ContextModel.START_TOKENS] + suggestions + \
                      [[tok] for tok in ContextModel.END_TOKENS]
        memory = [[MemoryItem(score=0.0, decoded=tuple())], [MemoryItem(score=0.0, decoded=tuple())]]
        for t in range(2, len(suggestions)):
            memory.append([])
            for i, word in enumerate(suggestions[t]):
                mx_score, pick_1, pick_2 = 0, 0, 0
                for j, suggestion_1 in enumerate(suggestions[t - 1]):
                    for k, suggestion_2 in enumerate(suggestions[t - 2]):
                        curr_score = memory[-3][k].score \
                            + self.model_dict.get((suggestion_2, suggestion_1), self.default_prob) \
                            + self.model_dict.get((suggestion_1, word), self.default_prob) \
                            + self.model_dict.get((suggestion_2, word), self.default_prob)
                        if curr_score > mx_score:
                            mx_score, pick_1, pick_2 = curr_score, j, k
                memory_item = MemoryItem(score=mx_score, decoded=memory[-3][pick_2].decoded + (pick_2, pick_1,))
                memory[-1].append(memory_item)
            memory = memory[1:]

        print(len(memory[-1][0].decoded))
        decoded = ' '.join([suggestions[t][i] for t, i in enumerate(memory[-1][0].decoded[-sent_word_count:],
                                                                    start=2)])
        # score = memory[-1][0].score
        return decoded

    def context_spell_correct(
            self,
            sentence: str,
            suggestions_dict: Dict[str, List[str]]
    ) -> (str, Dict[str, str]):
        """
        Get most probable suggestion for miss-spelled words from suggestions list
        Args:
            sentence (str): sentence
            suggestions_dict (dict): dict having miss-spelled word as key and list of suggestions as values
        Returns(str, Dict[str, str]): Tuple containing corrected string and word-to-word correction dictionary.
        """
        sentence_tokens = sentence.lower().split()
        possible_sent_list = []
        for each in sentence_tokens:
            if each in suggestions_dict:
                possible_sent_list.append(suggestions_dict[each])
            else:
                possible_sent_list.append([each])

        corrected_sent = self.get_most_probable_sentence(possible_sent_list)
        corrected_dict = self.get_corrected_words_map(corrected_sent, sentence)
        context_corrected_sentence = " " + sentence + " "
        for token, suggestion in corrected_dict.items():
            context_corrected_sentence = re.sub(" " + token + " ", " " + suggestion + " ", context_corrected_sentence)
        context_corrected_sentence = context_corrected_sentence.strip()
        return context_corrected_sentence, corrected_dict



import logging
import re
import os
import operator
import pickle
from collections import Counter
from pathlib import Path
from typing import Dict, List, Tuple, Any, Union

from spello.config import Config
from spello.context.context import ContextModel
from spello.phoneme.phoneme import PhonemeModel
from spello.settings import logger, loglevel
from spello.symspell.symspell import SymSpell
from spello.utils import get_clean_text

from spello.utils import mkdirs


ORIGINAL_TEXT = 'original_text'
CORRECTED_TEXT = 'spell_corrected_text'
CORRECTIONS_DICT = 'correction_dict'


class SpellCorrectionModel(object):
    """
    Model to find and correct miss-spelled words utter in text from two words corpus, domain and global dictionary.
    For each word in text which do not belong to either of domain or global dictionary, model considered it as
    miss-spelled word and try to find out most suitable suggestion first from domain dict, and then from global.

    Current model works on 4 key algorithms, which are
        1) Symspell Model - Suggest words based on simple edit distance.
        2) Phoneme Model - Suggest words based on Soundex code of word.
        3) Context Model -  Model to suggest most suitable word from suggestion list given by above two model
        4) Normalize Model -  Model to Normalize unigrams which also occurs as bigrams or trigrams to split form.
                              for example - Login to log in.
    """

    def __init__(
            self,
            language: str = 'en'
    ) -> None:
        self.language = language
        if language not in PhonemeModel.supported_languages + ['en']:
            raise Exception(f'"{language}" language is not yet supported')

        self.config = None
        self.set_default_config()
        self.symspell_model = SymSpell(config=self.config, script=self.language)
        self.phoneme_model = None
        if self.language in PhonemeModel.supported_languages:
            self.phoneme_model = PhonemeModel(config=self.config, script=self.language)
        self.context_model = None

    def set_default_config(self):
        self.config = Config

    def symspell_train(self, words_counter: Dict[str, int]) -> 'SymSpell':
        """
        Train symspell model
        Args:
            words_counter (dict): dict of word and their count
        Returns:
            (SymSpell)
        """
        self.symspell_model = SymSpell(config=self.config, script=self.language)
        self.symspell_model.create_dictionary_from_words(words_counter)
        return self.symspell_model

    def phoneme_train(self, words_counter: Dict[str, int]) -> 'PhonemeModel':
        """
        Train phoneme model
        Args:
            words_counter (dict): dict of word and their count
        Returns:
            (PhonemeModel)
        """
        if self.language not in PhonemeModel.supported_languages:
            return self.phoneme_model
        self.phoneme_model = PhonemeModel(config=self.config, script=self.language)
        self.phoneme_model.create_phoneme_dictionary_from_words(words_counter)
        return self.phoneme_model

    def context_train(self, texts: List[str]) -> 'ContextModel':
        """
        Train context model
        Args:
            texts (list): list of text
        Returns:
            (ContextModel)
        """
        self.context_model = ContextModel()
        self.context_model.create_model_dict(texts)
        return self.context_model

    def context_suggestion(self, text: str, suggestions_dict: Dict[str, List[str]]) -> Tuple[str, Dict[str, str]]:
        """
        Get context suggested text for given text and suggestions dict with misspelled words and their list
        of suggestions
        Args:
            text (str): text
            suggestions_dict: dict with misspelled words and their list of suggestions
        Returns:
            (str): contextually corrected text
            (dict): token level suggestion from context model

        Examples:
            >>> text = 'i wnt to pley kriket'
            >>> suggestions_dict = {'wnt': ['want', 'went', 'what'], 'pley': ['play'], 'kriket': ['cricket']}
            >>> SpellCorrectionModel().context_suggestion(text, suggestions_dict)
            (i want to play cricket, {'wnt': 'want', 'pley: 'play', 'kriket': 'cricket'})
        """
        if not self.context_model or not suggestions_dict:
            suggestions_dict = {key: value[0] for key, value in suggestions_dict.items()}
            return text, suggestions_dict
        text, context_corrections = self.context_model.context_spell_correct(text, suggestions_dict)
        return text, context_corrections

    def train(self, data: Union[List[str], Dict[str, int]], **kwargs):
        """
        Train all models of spellcorrection
        Args:
            data (list|dict): list of text or dict having word and their count
        Returns:
            None
        """

        logger.debug("Spello training started..")

        words_counter = {}
        if not isinstance(data, list) and not isinstance(data, dict):
            raise ValueError('Argument `data` should be either List[str] or Dict[str, int]')

        if isinstance(data, list):
            texts = [get_clean_text(text) for text in data]
            lower_case_texts = [text.lower() for text in texts]

            logger.debug("Context model training started ...")
            # Context model get trained only when list of text are given for training
            # train context model: find most probable correct word for given suggestions for each word in texts
            # based on context word
            self.context_train(lower_case_texts)

            tokens = " ".join(lower_case_texts).strip().split()

            words_counter = dict(Counter(tokens))

        elif isinstance(data, dict):
            words_counter = {word.lower(): count for word, count in data.items()}

        words_counter = {word: count for word, count in words_counter.items()
                         if self.config.min_length_for_spellcorrection}

        logger.debug("Symspell training started ...")
        # train symspell model: give suggestion based on edit distance
        self.symspell_train(words_counter)

        logger.debug("Phoneme training started ...")
        # train phoneme model: give suggestion for similar sounding words
        self.phoneme_train(words_counter)

        logger.debug("Spello training completed successfully ...")

    def _correct_word(self, word):
        """
        Suggest words for given word, follow below steps:
            - check if length of word is eligible for correction from min max length allowed from config,
                else return None
            - if word belong to domain word correction (created for words which occur less in domain words and are
                replaced with most similar words, either from domain or global), return suggested word from dict
            - if word exists in global dictionary, return None
            - if word exists in domain dictionary, return None
            - get suggestions from domain symspell & phoneme
            - if not suggestions from domain, get suggestions from global symspell and phoneme
            - filter top 5 suggestions
            - also add top 3 suggestions from global data up to min edit distance of above dict
            - return final suggestions
        Args:
            word (str): word to be corrected
        Returns:
            (list): list of suggested words
        """
        if ((self.config.min_length_for_spellcorrection > len(word)) or
                (len(word) > self.config.max_length_for_spellcorrection) or
                (re.search(r'\d', word))):
            return None

        phoneme = self.phoneme_model.spell_correct(word)
        if phoneme.is_correct:
            return None
        phoneme_suggestions = phoneme.suggestions

        symspell = self.symspell_model.spell_correct(word)
        if symspell.is_correct:
            return None
        symspell_suggestions = symspell.suggestions

        logger.debug(f"Symspell suggestions: {symspell_suggestions}")
        logger.debug(f"Phoneme suggestions: {phoneme_suggestions}")

        phoneme_suggestions = [(word, edit_distance, 0) for (word, edit_distance) in phoneme_suggestions]
        symspell_suggestions = [(word, edit_distance, 1) for (word, edit_distance) in symspell_suggestions]

        suggestions = list(phoneme_suggestions + symspell_suggestions)
        suggestions.sort(key=operator.itemgetter(1, 2))

        # filtering duplicate suggestions
        unique_suggestions = []
        for suggestion in suggestions:
            if suggestion[0] not in unique_suggestions:
                unique_suggestions.append(suggestion[0])

        suggestions = unique_suggestions

        return suggestions

    def spell_correct(self, text: str, verbose=0) -> Dict[str, Any]:
        """
        Get spell corrected text, and dict of token level suggestion
        Args:
            text (str):
            verbose (int): define verbose level
        Returns:
            (str): spell corrected text
            (dict): dict of token level suggestion

        Examples:
            >>> text = 'i wnt to play kricket'
            >>> SpellCorrectionModel().spell_correct(text)
            {
                'original_text': 'i wnt to play kricket',
                'corrected_text': 'i want to play cricket',
                'correction_dict': {'wnt': 'want', 'kricket': 'cricket'}
            }
        """
        levels = [logging.CRITICAL, logging.ERROR, logging.INFO, logging.DEBUG]
        verbosity = min(verbose, len(levels) - 1)
        with loglevel(levels[verbosity]):
            spellcorrection_result = {
                ORIGINAL_TEXT: text,
                CORRECTED_TEXT: text,
                CORRECTIONS_DICT: {}
            }

            suggestions_dict = {}
            clean_text = get_clean_text(text)
            tokens = clean_text.split()
            for token in tokens:
                lowercase_token = token.lower()
                token_suggestion = self._correct_word(lowercase_token)
                if token_suggestion:
                    suggestions_dict[lowercase_token] = token_suggestion

            logger.debug(f"Suggestions dict from phoneme and symspell are: {suggestions_dict}")

            context_corrected_text, context_corrections = self.context_suggestion(clean_text, suggestions_dict)

            logger.debug(f"text after context model: {context_corrected_text}")

            spellcorrection_result[CORRECTED_TEXT] = context_corrected_text
            spellcorrection_result[CORRECTIONS_DICT] = context_corrections

            logger.debug(f"Spell-correction Results {spellcorrection_result}")

        return spellcorrection_result

    def get_state(self) -> Dict[str, Any]:
        return {
            'language': self.language,
            'symspell_model': self.symspell_model,
            'phoneme_model': self.phoneme_model,
            'context_model': self.context_model,
            'config': self.config
        }

    def set_state(self, state_dict: Dict[str, Any]) -> None:
        self.language = state_dict['language']
        self.config = state_dict['config']
        self.symspell_model = state_dict['symspell_model']
        self.phoneme_model = state_dict['phoneme_model']
        self.context_model = state_dict['context_model']

    def load(
            self,
            model_path: str,
            **kwargs
    ) -> 'SpellCorrectionModel':
        model_state = pickle.load(open(model_path, 'rb'))
        self.set_state(model_state)
        return self

    def save(self, model_save_dir: str, **kwargs) -> str:
        save_dir = Path(model_save_dir)
        mkdirs(save_dir, err_if_already_exists=False)
        state = self.get_state()
        pickle.dump(state, open(save_dir / 'model.pkl', "wb"))
        path = os.path.join(save_dir, 'model.pkl')
        return path

